-- QuestJournalHook for MultiQuestTracker

local IGNORE_SCENE_RESTRICTION = true
local TRACKER = _G["MULTI_QUEST_TRACKER"]

local QuestJournalHook = ZO_CallbackObject:Subclass()

QuestJournalHook.UI = {}
QuestJournalHook.UI.Keyboard = {}
QuestJournalHook.UI.Gamepad = {}

function QuestJournalHook:New()

	local instance = ZO_CallbackObject.New( self )

	instance.UI.Keyboard.ui = QUEST_JOURNAL_KEYBOARD
	instance.UI.Keyboard.common = instance

	instance.UI.Gamepad.ui = QUEST_JOURNAL_GAMEPAD
	instance.UI.Gamepad.common = instance

	instance.manager = QUEST_JOURNAL_MANAGER

	instance:Initialize()

	return instance

end

-- T キーを押したときに QUEST_TRACKER が呼ぶ関数
-- 次に移るべきクエストフォーカス(Assist)をトラッカーに返すようだ
-- 以前(1年ほど前)は普通にできていたような？きっと気のせいだろう・・・
function QuestJournalHook.UI.Keyboard:GetNextSortedQuestForQuestIndex( questIndex )

	local firstNode = self.ui.questIndexToTreeNode[questIndex]
	if firstNode ~= nil
	then
		local currentNode = firstNode.nextNode
		repeat
			local index = currentNode.data.questIndex
			-- trackedなクエストのみサイクルする
			if GetIsTracked( TRACK_TYPE_QUEST, index )
			then
				return index
			end
			currentNode = currentNode.nextNode
		until currentNode == firstNode
	end

	return questIndex

end

function QuestJournalHook.UI.Gamepad:GetNextSortedQuestForQuestIndex( questIndex )

	local questMasterList = self.ui.questMasterList
	if questMasterList ~= nil
	then
		local function Inc( i )
			return ( i < #questMasterList ) and i + 1 or 1
		end
		for i = 1, #questMasterList
		do
			if questMasterList[i].questIndex == questIndex
			then
				local k = Inc( i )
				repeat
					local index = questMasterList[k].questIndex
					-- trackedなクエストのみサイクルする
					if GetIsTracked( TRACK_TYPE_QUEST, index )
					then
						return index
					end
					k = Inc( k )
				until k == i
				break
			end
		end
	end

	return questIndex

end

function QuestJournalHook.UI.Keyboard:IsJournalOpen()

	return SCENE_MANAGER:IsShowing( self.ui:GetSceneName() )

end

function QuestJournalHook.UI.Gamepad:IsJournalOpen()

	return SCENE_MANAGER:IsShowing( self.ui:GetSceneName() )

end

local function DoTrackingQuest( questIndex, tracked )

	if tracked
	then
		if GetIsTracked( TRACK_TYPE_QUEST, questIndex )
		then
			return true
		end
		if QUEST_TRACKER:GetNumTracked() + 1 > TRACKER:GetMaxNumTracked()
		then
			ZO_Alert( UI_ALERT_CATEGORY_ERROR, SOUNDS.NEGATIVE_CLICK, SI_MULTI_QUEST_TRACKER_MAXTRACKED )
			return false
		end
	end

	QUEST_TRACKER:SetTracked( questIndex, tracked )

	return true

end

function QuestJournalHook.UI.Keyboard:InitializeKeybindStripDescriptors()

	-- ジャーナルのキーガイド"Cycle Focused Quest"の出現条件を変更
	self.ui.keybindStripDescriptor[1].visible = function ()
		return ( QUEST_TRACKER:GetNumTracked() >= 2 )
	end

	-- 一時フォーカス時は消す
	local originalVisible2 = self.ui.keybindStripDescriptor[2].visible
	self.ui.keybindStripDescriptor[2].visible = function ()
		return ( originalVisible2() and self.temporaryFocusedIndex == nil )
	end

	local originalVisible3 = self.ui.keybindStripDescriptor[3].visible
	self.ui.keybindStripDescriptor[3].visible = function ()
		return ( originalVisible3() and self.temporaryFocusedIndex == nil )
	end

	local originalVisible4 = self.ui.keybindStripDescriptor[4].visible
	self.ui.keybindStripDescriptor[4].visible = function ()
		return ( originalVisible4() and self.temporaryFocusedIndex == nil )
	end

	self.questJournalkeybindStripDescriptor = 
	{
		alignment = KEYBIND_STRIP_ALIGN_CENTER,
		{
			name = GetString( SI_BINDING_NAME_TRACKQUEST_CURRENT_MAPZONE ),
			keybind = "UI_TRACKQUEST_CURRENT_MAPZONE",

			callback = function()
				MultiQuestTracker_TrackCurrentRegion()
			end,

			visible = function()
				return ( MultiQuestTracker_NumOfCurrentRegionQuests() > 0 and self.temporaryFocusedIndex == nil )
			end 
		}
	}

end

function QuestJournalHook.UI.Gamepad:InitializeKeybindStripDescriptors()

	self.questJournalkeybindStripDescriptor = 
	{
		alignment = KEYBIND_STRIP_ALIGN_LEFT,
		{
			name = GetString( SI_MULTI_QUEST_TRACKER_TRACKQUEST ),
			keybind = "UI_SHORTCUT_LEFT_STICK",

			callback = function ()
				local selectedQuestIndex = self.ui:GetSelectedQuestIndex()
				if selectedQuestIndex ~= nil
				then
					local isTracked = GetIsTracked( TRACK_TYPE_QUEST, selectedQuestIndex )
					if isTracked
					then
						self.trackedKeybindDescriptor.name = GetString( SI_MULTI_QUEST_TRACKER_UNTRACKQUEST )
					else
						self.trackedKeybindDescriptor.name = GetString( SI_MULTI_QUEST_TRACKER_TRACKQUEST )
					end

					local newTracked = ( not isTracked )
					if not newTracked and GetTrackedIsAssisted( TRACK_TYPE_QUEST, selectedQuestIndex )
					then
						QUEST_TRACKER:AssistNext( IGNORE_SCENE_RESTRICTION )
					end

					DoTrackingQuest( selectedQuestIndex, newTracked )

					self.ui:RefreshQuestCount()
					self.ui:RefreshQuestMasterList()
					self.ui:RefreshQuestList()

					KEYBIND_STRIP:UpdateKeybindButtonGroup( self.questJournalkeybindStripDescriptor )
				end
			end,

			visible = function ()
				return ( self.ui:GetSelectedQuestIndex() ~= nil )
			end,
		},
		{
			name = GetString( SI_BINDING_NAME_TRACKQUEST_CURRENT_MAPZONE ),
			keybind = "UI_SHORTCUT_RIGHT_STICK",

			callback = function()
				MultiQuestTracker_TrackCurrentRegion()
			end,

			visible = function()
				return ( MultiQuestTracker_NumOfCurrentRegionQuests() > 0 )
			end 
		}
	}
	
	self.trackedKeybindDescriptor = self.questJournalkeybindStripDescriptor[1]

end

local function KeyboardUI_EnableSelectionText( nodeTable, questIndex, isEnable )

	local node = nodeTable[questIndex]
	if node ~= nil
	then
		node.control:SetSelected( isEnable )
	end

end

function QuestJournalHook.UI.Keyboard:SetFocusIndex( questIndex )

	if self:IsJournalOpen() and self.temporaryFocusedIndex ~= nil
	then
		local enabled = ( questIndex == self.temporaryFocusedIndex )
		KeyboardUI_EnableSelectionText( self.ui.questIndexToTreeNode, self.temporaryFocusedIndex, enabled )

		self.temporaryFocusedIndex = nil
		self.suspendedFocusedIndex = nil

		KEYBIND_STRIP:UpdateKeybindButtonGroup( self.ui.keybindStripDescriptor )
		KEYBIND_STRIP:UpdateKeybindButtonGroup( self.questJournalkeybindStripDescriptor )
	end

	self.common.manager.focusedQuestIndex = questIndex

end

function QuestJournalHook.UI.Gamepad:SetFocusIndex( questIndex )

	if self:IsJournalOpen() and self.temporaryFocusedIndex ~= nil
	then
		self.temporaryFocusedIndex = nil
		self.suspendedFocusedIndex = nil
	end

	self.common.manager.focusedQuestIndex = questIndex

end

function QuestJournalHook.UI.Gamepad:FocusQuestWithIndex( questIndex )

	local prevFocusedIndex = self.common.manager:GetFocusedQuestIndex()

	self:SetFocusIndex( questIndex )

	self.ui:FireCallbacks( "QuestSelected", questIndex )

	if self:IsJournalOpen() and GetIsTracked( TRACK_TYPE_QUEST, questIndex )
	then
		QUEST_TRACKER:ForceAssist( questIndex )
	end

	self.ui:RefreshQuestMasterList()
	self.ui:RefreshQuestList()

	-- QUEST_TRACKER の "Show detail" コンテキストメニューから呼ばれた場合
	if SCENE_MANAGER:IsShowingBaseScene() and not GetTrackedIsAssisted( TRACK_TYPE_QUEST, questIndex )
	then
		-- 一時フォーカスモードに設定する
		-- ジャーナルのフォーカスが Not Assisted な場合は一時的なものとして扱う
		self.temporaryFocusedIndex = questIndex
		self.suspendedFocusedIndex = prevFocusedIndex
	end

end

function QuestJournalHook.UI.Keyboard:FocusQuestWithIndex( questIndex )

	local node = self.ui.questIndexToTreeNode[questIndex]
	if node ~= nil
	then
		local prevFocusedIndex = self.common.manager:GetFocusedQuestIndex()

		self:SetFocusIndex( questIndex )

		local orgClass = getmetatable( self.ui )
		orgClass.FocusQuestWithIndex( self.ui, questIndex )

		-- QUEST_TRACKER の "Show detail" コンテキストメニューから呼ばれた場合
		if SCENE_MANAGER:IsShowingBaseScene() and not GetTrackedIsAssisted( TRACK_TYPE_QUEST, questIndex )
		then
			-- 一時フォーカスモードに設定する
			-- ジャーナルのフォーカスが Not Assisted な場合は一時的なものとして扱う
			self.temporaryFocusedIndex = questIndex
			self.suspendedFocusedIndex = prevFocusedIndex
		end
	end

end

function QuestJournalHook.UI.Keyboard:RefreshDetails()

	local selected = self.ui:GetSelectedQuestData()
	if selected ~= nil
	then
		if not GetTrackedIsAssisted( TRACK_TYPE_QUEST, selected.questIndex )
			and selected.displayType == INSTANCE_DISPLAY_TYPE_NONE 
		then
			self.ui.questIcon:SetHidden( true )
		end
	end

end

local function KeyboardUI_UpdateQuestEntryIcon( questJournal, control, displayType, isAssisted )

	if isAssisted
	then
		control:SetTexture( "EsoUI/Art/Journal/journal_Quest_Selected.dds" )
		control:SetAlpha( 1.00 )
		control:SetHidden( false )
	else
		control:SetTexture( questJournal:GetIconTexture( displayType ) )
		if displayType == INSTANCE_DISPLAY_TYPE_NONE
		then
			control:SetHidden( true )
		else
			-- selected 以外にも何かアイコンが設定される模様
			control:SetAlpha( 0.50 )
			control:SetHidden( false )
		end
	end

end

function QuestJournalHook.UI.Keyboard:SetAssistedIcon( questIndex, isAssisted )

	local node = self.ui.questIndexToTreeNode[questIndex]
	if node ~= nil
	then
		local control = GetControl( node.control, "Icon" )
		local displayType = node.data.displayType

		KeyboardUI_UpdateQuestEntryIcon( self.ui, control, displayType, isAssisted )

		-- Multi-Quest Tracker REではassistedなクエストにのみjournal_Quest_Selected.ddsを置く
		local assistedData = QUEST_TRACKER.assistedData
		if assistedData ~= nil
		then
			local assistedIndex = assistedData:GetJournalIndex()
			if assistedIndex ~= questIndex
			then
				local node = self.ui.questIndexToTreeNode[assistedIndex]
				if node ~= nil
				then
					local control = GetControl( node.control, "Icon" )
					local displayType = node.data.displayType

					KeyboardUI_UpdateQuestEntryIcon( self.ui, control, displayType, true )
				end
			end
		end
	end

end

function QuestJournalHook.UI.Keyboard:OnAssistChanged( unassistedData, assistedData )

	-- Workaround: ジャーナルの "Cycled Focused Quest" が動作しない問題を回避
	if assistedData ~= nil
	then
		local assistedIndex = assistedData:GetJournalIndex()
		if assistedIndex ~= self.ui:GetSelectedQuestIndex()
		then
			self.ui:FocusQuestWithIndex( assistedIndex )
		end
	end

	if unassistedData ~= nil
	then
		self:SetAssistedIcon( unassistedData:GetJournalIndex(), false )
	end

end

local JOURNAL_CHECKBOX_NAME = "MultiQuestTrackerCheckBox"

local function KeyboardUI_SetEnableCheckBoxes( questJournal )

	local isEnabled = ( QUEST_TRACKER:GetNumTracked() < TRACKER:GetMaxNumTracked() )
	for _, node in pairs( questJournal.questIndexToTreeNode )
	do
		local checkbox = GetControl( node.control, JOURNAL_CHECKBOX_NAME )
		if checkbox ~= nil and not ZO_CheckButton_IsChecked( checkbox )
		then
			checkbox:SetEnabled( isEnabled )
		end
	end

	questJournal:RefreshQuestCount()

end

function QuestJournalHook.UI.Keyboard:RefreshQuestCount()

	local text = zo_strformat( SI_MULTI_QUEST_TRACKER_QUESTCOUNT, self.ui.questCount:GetText(), QUEST_TRACKER:GetNumTracked(), TRACKER:GetMaxNumTracked() )
	self.ui.questCount:SetText( text )

end

function QuestJournalHook.UI.Keyboard:ReinitializeQuestList()

	-- QUEST_TRACKERのset系はジャーナル表示中のみ呼び出すようにする
	local function DoCallQuestTrackerMethod( methodName, ... )
		if self:IsJournalOpen()
		then
			local questTrackerFunction = QUEST_TRACKER[methodName]
			questTrackerFunction( QUEST_TRACKER, ... )
		end
	end

	-- ジャーナルのチェックボックスのハンドラ
	local function OnToggleCheckButton( control )
		local parent = control:GetParent()
		local questIndex = parent.questIndex

		local isChecked = ZO_CheckButton_IsChecked( control )
		local isSelected = parent:IsSelected()
		local isAssisted = GetTrackedIsAssisted( TRACK_TYPE_QUEST, questIndex )
		local isTrackerEmpty = ( QUEST_TRACKER:GetNumTracked() == 0 )

		if isChecked and ( isTrackerEmpty or isSelected )
		then
			DoCallQuestTrackerMethod( "ForceAssist", questIndex )
			self:SetAssistedIcon( questIndex, true )
			isAssisted = true
		elseif not isChecked and isAssisted
		then
			DoCallQuestTrackerMethod( "AssistNext", IGNORE_SCENE_RESTRICTION )
			self:SetAssistedIcon( questIndex, false )
			isAssisted = false
		end

		DoCallQuestTrackerMethod( "SetTracked", questIndex, isChecked )

		if questIndex == self.common.manager:GetFocusedQuestIndex()
		then
			local displayType = GetJournalInstanceDisplayType( questIndex )
			local detailsIconVisible = ( isAssisted or displayType ~= INSTANCE_DISPLAY_TYPE_NONE )
			self.ui.questIcon:SetHidden( ( not detailsIconVisible ) )
		end

		KeyboardUI_SetEnableCheckBoxes( self.ui )

		KEYBIND_STRIP:UpdateKeybindButtonGroup( self.ui.keybindStripDescriptor )
		KEYBIND_STRIP:UpdateKeybindButtonGroup( self.questJournalkeybindStripDescriptor )
	end

	local templates = self.ui.navigationTree.templateInfo["ZO_QuestJournalNavigationEntry"]
	local originalSetup = templates.setupFunction

	-- ジャーナルのクエストリストの初期化関係
	local onSetup = function ( node, control, data, open )
		originalSetup( node, control, data, open )

		local checkBox = GetControl( control, JOURNAL_CHECKBOX_NAME )
		if checkBox == nil
		then
			local name = control:GetName() .. JOURNAL_CHECKBOX_NAME

			checkBox = WINDOW_MANAGER:CreateControlFromVirtual( name, control, "ZO_CheckButton" )
			checkBox:SetAnchor( RIGHT, control, LEFT, -25, 0 )
		end

		ZO_CheckButton_SetToggleFunction( checkBox, nil ) -- 無用なイベント処理を防ぐため
		ZO_CheckButton_SetCheckState( checkBox, GetIsTracked( TRACK_TYPE_QUEST, data.questIndex ) )
		ZO_CheckButton_SetToggleFunction( checkBox, OnToggleCheckButton )

		checkBox:SetHidden( false )
	end

	-- ジャーナルのクエストリストの選択処理ハンドラ
	local onSelection = function ( control, data, selected, reselectingDuringRebuild )
		local questIndex = data.questIndex

		self.ui:FireCallbacks( "QuestSelected", questIndex )

		control:SetSelected( selected )

		if selected and not reselectingDuringRebuild
		then
			if GetIsTracked( TRACK_TYPE_QUEST, questIndex )
			then
				local checkBox = GetControl( control, JOURNAL_CHECKBOX_NAME )
				if not ZO_CheckButton_IsChecked( checkBox )
				then
					ZO_CheckButton_SetChecked( checkBox )
				end
				DoCallQuestTrackerMethod( "ForceAssist", questIndex )
			end
			self:SetFocusIndex( questIndex )
			self.ui:RefreshDetails()
		end

		GetControl( control, "Icon" ).selected = selected

		local isAssisted = ( selected and GetTrackedIsAssisted( TRACK_TYPE_QUEST, questIndex ) )
		self:SetAssistedIcon( questIndex, isAssisted )
	end

	templates.selectionFunction = onSelection
	templates.setupFunction = onSetup

	-- 再設定したので呼び直す
	self.ui:RefreshQuestList()

end

local QUEST_TEMPLATE = "ZO_QuestJournal_Gamepad_MenuEntryTemplate"
local SELECTED_QUEST_TEXTURE = "EsoUI/Art/Journal/Gamepad/gp_trackedQuestIcon.dds"

function QuestJournalHook.UI.Gamepad:ReinitializeQuestList()

--	local onSetup = function ( control, data, selected, selectedDuringRebuild, enabled, activated )
	local onSetup = function ( ... )
		ZO_SharedGamepadEntry_OnSetup( ... )

		local control, data = select( 1, ... )

		control.label:SetColor( GetConColor( data.level ) )

		local controlAlpha = 0.35
		local hidden = true

		if GetIsTracked( TRACK_TYPE_QUEST, data.questIndex )
		then
			hidden = ( not GetTrackedIsAssisted( TRACK_TYPE_QUEST, data.questIndex ) )
			controlAlpha = 1.00
		end

		control.selectedIcon:SetTexture( SELECTED_QUEST_TEXTURE )
		control.selectedIcon:SetHidden( hidden )

		control:SetAlpha( controlAlpha )
	end

	local templates = self.ui.questList.dataTypes[QUEST_TEMPLATE]
	templates.setupFunction = onSetup

	local templatesWithHeader = self.ui.questList.dataTypes[QUEST_TEMPLATE .. "WithHeader"]
	templatesWithHeader.setupFunction = onSetup

    local selectedIcon = zo_iconFormat( "EsoUI/Art/Journal/Gamepad/gp_trackedQuestIcon.dds", 32, 32 )
	self.ui.contentHeaderData.titleText = function ()
		local data = self.ui:GetSelectedQuestData()
		if data ~= nil
		then
			local instanceDisplayType = GetJournalInstanceDisplayType( data.questIndex )
			local color = ZO_ColorDef:New( GetConColor( data.level ) )
			local text = color:Colorize( GetJournalQuestName( data.questIndex ) )
			local icon = zo_iconFormat( self.ui:GetIconTexture( instanceDisplayType ), 48, 48 )

			if self.common.manager:GetFocusedQuestIndex() == data.questIndex 
				and GetTrackedIsAssisted( TRACK_TYPE_QUEST, data.questIndex )
			then
				return zo_strformat( SI_GAMEPAD_SELECTED_QUEST_JOURNAL_QUEST_NAME_FORMAT, selectedIcon, icon, text )
			else
				return zo_strformat( SI_GAMEPAD_QUEST_JOURNAL_QUEST_NAME_FORMAT, icon, text )
			end
		end
	end

	local orgData1Text = self.ui.headerData.data1Text
	self.ui.headerData.data1Text = function ()
		return zo_strformat( SI_MULTI_QUEST_TRACKER_QUESTCOUNT, orgData1Text(), QUEST_TRACKER:GetNumTracked(), TRACKER:GetMaxNumTracked() )
	end

	-- 再設定したので呼び直す
	self.ui:RefreshQuestCount()
	self.ui:RefreshQuestList()

end

function QuestJournalHook.UI.Keyboard:AfterQuestAdded( questIndex )

	self.ui:RefreshQuestCount()
	self.ui:RefreshQuestList()

end

function QuestJournalHook.UI.Gamepad:AfterQuestAdded( questIndex )

	self.ui:RefreshQuestCount()
	self.ui:RefreshQuestList()

end

function QuestJournalHook.UI.Keyboard:AfterQuestRemoved()

	self.ui:RefreshQuestCount()
	self.ui:RefreshQuestList()

	local isOpenCategory = false

	-- 開いているカテゴリを探す
	local rootNode = self.ui.navigationTree.rootNode
	if rootNode ~= nil and rootNode.children ~= nil
	then
		for _, categoryNode in ipairs( rootNode.children )
		do
			isOpenCategory = categoryNode:IsOpen()
			if isOpenCategory
			then
				break
			end
		end
	end

	-- 開いていたカテゴリ内のクエストが全て無くなった場合
	if not isOpenCategory
	then
		-- 選択中のクエストのカテゴリを開く
		local selectedNode = self.ui.navigationTree.selectedNode
		if selectedNode ~= nil
		then
			self.ui.navigationTree:ToggleNode( selectedNode:GetParent() )
		end
	end

	KEYBIND_STRIP:UpdateKeybindButtonGroup( self.ui.keybindStripDescriptor )
	KEYBIND_STRIP:UpdateKeybindButtonGroup( self.questJournalkeybindStripDescriptor )

end

function QuestJournalHook.UI.Gamepad:AfterQuestRemoved()

	self.ui:RefreshQuestCount()
	self.ui:RefreshQuestList()

	KEYBIND_STRIP:UpdateKeybindButtonGroup( self.ui.mainKeybindStripDescriptor )
	KEYBIND_STRIP:UpdateKeybindButtonGroup( self.questJournalkeybindStripDescriptor )

end

function QuestJournalHook.UI.Gamepad:RefreshDetails()

	if not self:IsJournalOpen()
	then
		return 
	end

	local selectedQuestIndex = self.ui:GetSelectedQuestIndex()
	if selectedQuestIndex ~= nil
	then
		if GetIsTracked( TRACK_TYPE_QUEST, selectedQuestIndex )
		then
			self.trackedKeybindDescriptor.name = GetString( SI_MULTI_QUEST_TRACKER_UNTRACKQUEST )
		else
			self.trackedKeybindDescriptor.name = GetString( SI_MULTI_QUEST_TRACKER_TRACKQUEST )
		end
		KEYBIND_STRIP:UpdateKeybindButtonGroup( self.questJournalkeybindStripDescriptor )
	end

end

function QuestJournalHook.UI.Keyboard:RefreshQuestList()

	KeyboardUI_SetEnableCheckBoxes( self.ui )

end

-- ZO_PreHook() とは違って関数を完全に置き換えてしまう
local function HookMethod( hook, object, methodName, hookFunction )

	local objectMethod = object[methodName]
	if objectMethod ~= nil and type( objectMethod ) == "function"
	then
		object[methodName] = function ( ... )
			return hookFunction( hook, select( 2, ... ) ) -- 1はobject
		end
    end

end

-- ZO_PreHook() の逆バージョン
-- hookFunctionが後から呼ばれる
local function PostHookMethod( hook, object, methodName, hookFunction )

	local objectMethod = object[methodName]
	if objectMethod ~= nil and type( objectMethod ) == "function"
	then
		object[methodName] = function ( ... )
			local args = { ... }

			local result = objectMethod( unpack( args ) )

			table.remove( args, 1 ) -- 1番目はobject
			table.insert( args, result ) -- 最後にオリジナルの関数の実行結果を格納

			return hookFunction( hook, unpack( args ) )
		end
    end

end

function QuestJournalHook:GetTrackerStateChangeHandler()

	-- チェックをON/OFFするたびに毎回 InitialTrackingUpdate() が呼ばれるのは
	-- 激しく無駄なので、クエストトラッカーへの反映は、ジャーナルを閉じるときにまとめて行う
	local function OnQuestTrackingStateChanged()
		if not self.isTrackedChanged
		then
			self.isTrackedChanged = true
		end
	end

	return OnQuestTrackingStateChanged

end

function QuestJournalHook.UI.Keyboard:ReinitializeMap()

	local mapQuestsControl = WORLD_MAP_QUESTS.control

	self.mapQuestsKeybindStripDescriptor =
	{
		alignment = KEYBIND_STRIP_ALIGN_CENTER,
		{
			name = GetString( SI_BINDING_NAME_TRACKQUEST_CURRENT_MAP ),
			keybind = "UI_TRACKQUEST_CURRENT_MAPZONE",

			callback = function()
				MultiQuestTracker_TrackCurrentMap()
			end,

			visible = function()
				return ( MultiQuestTracker_NumOfCurrentMapQuests() > 0 )
			end
		}
	}

--[[
	HookMethod( self, WORLD_MAP_QUESTS, "QuestHeader_OnClicked",
		function ( _, header, button )
			if button == MOUSE_BUTTON_INDEX_LEFT
			then
				if DoTrackingQuest( header.data.questIndex, true )
				then
					local orgClass = getmetatable( WORLD_MAP_QUESTS )
					orgClass.QuestHeader_OnClicked( WORLD_MAP_QUESTS, header, button )
				end
			end
		end )
--]]

	WORLD_MAP_QUESTS_FRAGMENT:RegisterCallback( "StateChange", 
		function ( oldState, newState )
			if newState == SCENE_SHOWING
			then
				KEYBIND_STRIP:AddKeybindButtonGroup( self.mapQuestsKeybindStripDescriptor )
			elseif newState == SCENE_HIDDEN
			then
				KEYBIND_STRIP:RemoveKeybindButtonGroup( self.mapQuestsKeybindStripDescriptor )
			end
		end )

	CALLBACK_MANAGER:RegisterCallback( "OnWorldMapQuestsDataRefresh",
		function ()
			KEYBIND_STRIP:UpdateKeybindButtonGroup( self.mapQuestsKeybindStripDescriptor )
		end )

end

function QuestJournalHook.UI.Keyboard:Initialize()

	local onQuestTrackingStateChanged = self.common:GetTrackerStateChangeHandler()

	local onStateChangeJournal = function ( oldState, newState )
		if newState == SCENE_SHOWING
		then
			-- 100012からJournalのSCENE_SHOWINGでFocusQuestWithIndex()が呼ばれるようになったので
			if self.temporaryFocusedIndex ~= nil
			then
				-- 実際の選択状態にすると色々不整合が出るので"見せかけの選択状態"にする
				self.ui.navigationTree:ClearSelectedNode()
				KeyboardUI_EnableSelectionText( self.ui.questIndexToTreeNode, self.temporaryFocusedIndex, true )
			end

			QUEST_TRACKER:RegisterCallback( "QuestTrackerTrackingStateChanged", onQuestTrackingStateChanged )

			KEYBIND_STRIP:AddKeybindButtonGroup( self.questJournalkeybindStripDescriptor )
		elseif newState == SCENE_HIDDEN
		then
			-- 一時的なジャーナルのフォーカスを破棄する
			if self.temporaryFocusedIndex ~= nil
				and self.suspendedFocusedIndex ~= self.common.manager:GetFocusedQuestIndex()
			then
				if self.suspendedFocusedIndex ~= nil
					and IsValidQuestIndex( self.suspendedFocusedIndex )
				then
					-- 以前のフォーカスに戻す
					KeyboardUI_EnableSelectionText( self.ui.questIndexToTreeNode, self.temporaryFocusedIndex, false )
					self.ui:FocusQuestWithIndex( self.suspendedFocusedIndex )
				end
				self.temporaryFocusedIndex = nil
				self.suspendedFocusedIndex = nil
			end

			QUEST_TRACKER:UnregisterCallback( "QuestTrackerTrackingStateChanged", onQuestTrackingStateChanged )

			KEYBIND_STRIP:RemoveKeybindButtonGroup( self.questJournalkeybindStripDescriptor )

			if self.common.isTrackedChanged
			then
				TRACKER:InitialTrackingUpdate()
				self.common.isTrackedChanged = false
			end
		end
	end

	QUEST_JOURNAL_SCENE:RegisterCallback( "StateChange", onStateChangeJournal )

	HookMethod( self, self.ui, "GetNextSortedQuestForQuestIndex", self.GetNextSortedQuestForQuestIndex )
	HookMethod( self, self.ui, "FocusQuestWithIndex", self.FocusQuestWithIndex )
	PostHookMethod( self, self.ui, "InitializeKeybindStripDescriptors", self.InitializeKeybindStripDescriptors )
	PostHookMethod( self, self.ui, "RefreshDetails", self.RefreshDetails )
	PostHookMethod( self, self.ui, "RefreshQuestList", self.RefreshQuestList )
	PostHookMethod( self, self.ui, "RefreshQuestCount", self.RefreshQuestCount )

	QUEST_TRACKER:RegisterCallback( "QuestTrackerAssistStateChanged",
		function ( unassistedData, assistedData )
			self:OnAssistChanged( unassistedData, assistedData )
		end )

	-- 再設定
	self.ui:InitializeKeybindStripDescriptors()
	self:ReinitializeQuestList()
	self:ReinitializeMap()

	-- EVENT_QUEST_LIST_UPDATEDはMULTI_QUEST_TRACKER側で処理する
	-- (trackedのリセットが起こるため)
	self.ui.control:UnregisterForEvent( EVENT_QUEST_LIST_UPDATED )

	-- On系関数の追加処理
	-- QUEST_TRACKERの処理を通った後に実行される
	-- (一部QUEST_TRACKERのデータに依存している部分があるため)
	self.common:RegisterCallback( "AfterQuestAdded", function ( ... ) self:AfterQuestAdded( ... ) end )
	self.common:RegisterCallback( "AfterQuestRemoved", function ( ... ) self:AfterQuestRemoved( ... ) end )

end

function QuestJournalHook.UI.Gamepad:ReinitializeMap()

	local mapQuestsControl = GAMEPAD_WORLD_MAP_QUESTS.control

	self.mapQuestsKeybindStripDescriptor =
	{
		alignment = KEYBIND_STRIP_ALIGN_LEFT,
		{
			name = GetString( SI_BINDING_NAME_TRACKQUEST_CURRENT_MAP ),
			keybind = "UI_SHORTCUT_RIGHT_STICK",

			callback = function()
				MultiQuestTracker_TrackCurrentMap()
				GAMEPAD_WORLD_MAP_QUESTS:RefreshHeaders()
			end,

			visible = function()
				return ( MultiQuestTracker_NumOfCurrentMapQuests() > 0 )
			end
		}
	}

--[[
	local orgMapDescriptorFunction1 = GAMEPAD_WORLD_MAP_QUESTS.keybindStripDescriptor[1].callback
	GAMEPAD_WORLD_MAP_QUESTS.keybindStripDescriptor[1].callback = function ()
		local data = GAMEPAD_WORLD_MAP_QUESTS.questList:GetTargetData()
		if data ~= nil
		then
			if DoTrackingQuest( data.questInfo.questIndex, true )
			then
				orgMapDescriptorFunction1()
			end
		end
	end
--]]

	GAMEPAD_WORLD_MAP_QUESTS_FRAGMENT:RegisterCallback( "StateChange", 
		function ( oldState, newState )
			if newState == SCENE_SHOWING
			then
				KEYBIND_STRIP:AddKeybindButtonGroup( self.mapQuestsKeybindStripDescriptor )
			elseif newState == SCENE_HIDDEN
			then
				KEYBIND_STRIP:RemoveKeybindButtonGroup( self.mapQuestsKeybindStripDescriptor )
			end
		end )

	CALLBACK_MANAGER:RegisterCallback( "OnWorldMapQuestsDataRefresh",
		function ()
			KEYBIND_STRIP:UpdateKeybindButtonGroup( self.mapQuestsKeybindStripDescriptor )
		end )

end

function QuestJournalHook.UI.Gamepad:Initialize()

	local onQuestTrackingStateChanged = self.common:GetTrackerStateChangeHandler()

	local onStateChangeJournal = function ( oldState, newState )
		if newState == SCENE_SHOWING
		then
			QUEST_TRACKER:RegisterCallback( "QuestTrackerTrackingStateChanged", onQuestTrackingStateChanged )

			KEYBIND_STRIP:AddKeybindButtonGroup( self.questJournalkeybindStripDescriptor )
		elseif newState == SCENE_HIDDEN
		then
			-- 一時的なジャーナルのフォーカスを破棄する
			if self.temporaryFocusedIndex ~= nil
				and self.suspendedFocusedIndex ~= self.common.manager:GetFocusedQuestIndex()
			then
				if self.suspendedFocusedIndex ~= nil
					and IsValidQuestIndex( self.suspendedFocusedIndex )
				then
					-- 以前のフォーカスに戻す
					self.ui:FocusQuestWithIndex( self.suspendedFocusedIndex )
				end
				self.temporaryFocusedIndex = nil
				self.suspendedFocusedIndex = nil
			end

			QUEST_TRACKER:UnregisterCallback( "QuestTrackerTrackingStateChanged", onQuestTrackingStateChanged )

			if self.common.isTrackedChanged
			then
				TRACKER:InitialTrackingUpdate()
				self.common.isTrackedChanged = false
			end

			KEYBIND_STRIP:RemoveKeybindButtonGroup( self.questJournalkeybindStripDescriptor )
		end
	end

	GAMEPAD_QUEST_JOURNAL_ROOT_SCENE:RegisterCallback( "StateChange", onStateChangeJournal )

	HookMethod( self, self.ui, "GetNextSortedQuestForQuestIndex", self.GetNextSortedQuestForQuestIndex )
	HookMethod( self, self.ui, "FocusQuestWithIndex", self.FocusQuestWithIndex )
	PostHookMethod( self, self.ui, "RefreshDetails", self.RefreshDetails )
	PostHookMethod( self, self.ui, "InitializeKeybindStripDescriptors", self.InitializeKeybindStripDescriptors )

	-- 再設定
	self.ui:InitializeKeybindStripDescriptors()
	self:ReinitializeQuestList()
	self:ReinitializeMap()

	-- EVENT_QUEST_LIST_UPDATEDはMULTI_QUEST_TRACKER側で処理する
	-- (trackedのリセットが起こるため)
	self.ui.control:UnregisterForEvent( EVENT_QUEST_LIST_UPDATED )

	-- On系関数の追加処理
	-- QUEST_TRACKERの処理を通った後に実行される
	-- (一部QUEST_TRACKERのデータに依存している部分があるため)
	self.common:RegisterCallback( "AfterQuestAdded", function ( ... ) self:AfterQuestAdded( ... ) end )
	self.common:RegisterCallback( "AfterQuestRemoved", function ( ... ) self:AfterQuestRemoved( ... ) end )

end

local function GetJournalSceneName()

	return SYSTEMS:GetObject( "questJournal" ):GetSceneName()

end

function QuestJournalHook:Initialize()

	for _, e in pairs( self.UI )
	do
		e:Initialize()
	end

	local onQuestTrackingStateChanged = self:GetTrackerStateChangeHandler()

	WORLD_MAP_FRAGMENT:RegisterCallback( "StateChange", 
		function ( oldState, newState )
			if newState == SCENE_SHOWING
			then
				QUEST_TRACKER:RegisterCallback( "QuestTrackerTrackingStateChanged", onQuestTrackingStateChanged )
			elseif newState == SCENE_HIDDEN
			then
				QUEST_TRACKER:UnregisterCallback( "QuestTrackerTrackingStateChanged", onQuestTrackingStateChanged )
				if self.isTrackedChanged
				then
					TRACKER:InitialTrackingUpdate()
					MultiQuestTracker_UpdateQuestJournal()
					self.isTrackedChanged = false
				end
			end
		end )

--[[
	-- クエストのPINハンドラにTrackerが一杯になった時の処理を挿入
	local questPinClickHandler = ZO_MapPin.PIN_CLICK_HANDLERS[1][MAP_PIN_TYPE_TRACKED_QUEST_CONDITION][1]
	-- クリック時のアクション
	local orgQuestPinClickHandlerFunc = questPinClickHandler.callback
	questPinClickHandler.callback = function ( pin )
		local questIndex = pin:GetQuestIndex()
		if DoTrackingQuest( questIndex, true )
		then
			orgQuestPinClickHandlerFunc( pin )
		end
	end
	-- 有効にするか否か
	local orgQuestPinClickHandlerShow = questPinClickHandler.show
	questPinClickHandler.show = function ( pin )
		local questIndex = pin:GetQuestIndex()
		if questIndex == -1
			or ( not GetIsTracked( TRACK_TYPE_QUEST, questIndex )
				and QUEST_TRACKER:GetNumTracked() >= TRACKER:GetMaxNumTracked() )
		then
			return false
		end
		return orgQuestPinClickHandlerShow( pin )
	end
--]]

	-- On系関数の追加処理
	-- QUEST_TRACKERの処理を通った後に実行される
	-- (一部QUEST_TRACKERのデータに依存している部分があるため)
	PostHookMethod( self, QUEST_TRACKER, "OnQuestAdded",
		function ( ... )
			self:FireCallbacks( "OnQuestAdded", ... )
			self:FireCallbacks( "AfterQuestAdded", ... )
		end )
	PostHookMethod( self, QUEST_TRACKER, "OnQuestRemoved",
		function ( ... )
			self:FireCallbacks( "OnQuestRemoved", ... )
			self:FireCallbacks( "AfterQuestRemoved", ... )
		end )

	-- QUEST_TRACKERのメンバ関数改造
	-- フルの基準をMULTI_QUEST_TRACKERの最大件数に変える
	HookMethod( self, QUEST_TRACKER, "IsFull",
		function ( ... )
			return ( QUEST_TRACKER:GetNumTracked() >= TRACKER:GetMaxNumTracked() )
		end )
	-- クエスト・トラッカーが満タンであっても先頭のクエストをUntrackedにして強制的に追加する
	HookMethod( self, QUEST_TRACKER, "ForceAssist",
		function ( _, questIndex )
			if not GetIsTracked( TRACK_TYPE_QUEST, questIndex )
			then
				if QUEST_TRACKER:IsFull()
				then
					local firstTrackedQuestIndex = TRACKER.tracked[1]:GetJournalIndex()
					QUEST_TRACKER:StopTracking( TRACK_TYPE_QUEST, firstTrackedQuestIndex )
				end
				QUEST_TRACKER:BeginTracking( TRACK_TYPE_QUEST, questIndex )
				if not SCENE_MANAGER:IsShowing( GetJournalSceneName() ) and not WORLD_MAP_FRAGMENT:IsShowing()
				then
					TRACKER:InitialTrackingUpdate()
					MultiQuestTracker_UpdateQuestJournal()
				end
			end
			local header = QUEST_TRACKER:GetHeaderForIndex( TRACK_TYPE_QUEST, questIndex )
			if header ~= nil
			then
				QUEST_TRACKER:SetAssisted( header, true )
			end
		end )

end

local QUEST_JOURNAL_HOOK = nil

function MultiQuestTracker_UpdateQuestJournal()

	local questJournal = SYSTEMS:GetObject( "questJournal" )
	questJournal:OnQuestsUpdated()

end

function MultiQuestTracker_SetQuestJournalHookCallbackHandler( callbackEventName, handler )

	if QUEST_JOURNAL_HOOK ~= nil
	then
		QUEST_JOURNAL_HOOK:RegisterCallback( callbackEventName, handler )
	end

end

function MultiQuestTracker_SetupQuestJournalHook()

	if QUEST_JOURNAL_HOOK == nil
	then
		QUEST_JOURNAL_HOOK = QuestJournalHook:New()
	end

end
