local ADDON_VERSION = "1.1.13.4"
local ADDON_SAVE_VERSION = "1.1.6"

local ADDON_SAVE_VARIABLES = "MultiQuestTracker_SavedVariables"

local MY_ADDON_INFO = {}

local function SplitVersion( verStr )

	local result = {}

	for v in string.gmatch( verStr, "%d+" )
	do
		table.insert( result, v )
	end

	return result

end

function MultiQuestTracker_ToVersionNumber( verStr )

	local ver = SplitVersion( verStr )

	return tonumber( string.format( "%d%d%d%02d", ver[1], ver[2], ver[3], ( ver[4] or 0 ) ) )

end

function MultiQuestTracker_GetVersion()

	return MultiQuestTracker_ToVersionNumber( MY_ADDON_INFO.addonVersion )

end

function MultiQuestTracker_GetVersionDisplayFormat()

	local stages =
	{
		[0] = "-pre",
		[1] = "-beta",
		[2] = "-rc",
		[3] = "", -- final version
	}

	local ver = SplitVersion( MY_ADDON_INFO.addonVersion )
	local stage = tonumber( ver[4] )

	ver[4] = stages[stage]
	if stage > 3
	then
		ver[4] = string.format( "-hotfix_%02d", ( stage - 3 ) )
	end

	return string.format( "%s.%s.%s%s", ver[1], ver[2], ver[3], ver[4] )

end

function MultiQuestTracker_GetAddonName()

	return MY_ADDON_INFO.name or "MultiQuestTrackerRE"

end

function MultiQuestTracker_GetAddonInfo( name )

	return MY_ADDON_INFO[name]

end

function MultiQuestTracker_AddonInitialize()

	local addonManager = GetAddOnManager()

	for i = 1, addonManager:GetNumAddOns()
	do
		local name, title, author, description, enabled, state, isOutOfDate = addonManager:GetAddOnInfo( i )
		if name == MultiQuestTracker_GetAddonName()
		then
			MY_ADDON_INFO = 
			{
				name = name,
				title = title,
				author = author,
				description = description,
				enabled = enabled,
				state = state,
				isOutOfDate = isOutOfDate,
				addonVersion = ADDON_VERSION,
				saveVersion = ADDON_SAVE_VERSION,
				saveVariables = ADDON_SAVE_VARIABLES,
			}
			break
		end
	end

end
